//
//  OCVCanvas.java
//  InfiniteCanvas
//
//  Created by Markus Mueller <markus@mindnode.com>.
//	http://www.infinitecanvas.com/
//  Copyright (c) 2003-2005 Markus Mueller. All rights reserved.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 

import java.awt.*;
import java.awt.geom.*;
import java.lang.Math;
import java.util.*;
import javax.swing.*;


public class OCVCanvas extends JComponent {
	// loading
	Image splashImage = null;
	Color loadingColor = Color.white;
	
	// general
	OCVDocument document = null;
	Vector previousVector = null;

	// canvas state
	int edge = 1;
	public Rectangle2D.Double canvasRect;
	float canvasRotation = 0;
	OCVRectObject activeFrame = null;

	
	// animation stuff
	int animationType = 0;
	float animationFadeRate = 0;
	OCVNavTrailObject animationCurrentPath = null;
	int times, animationFactor, segment;
	double dX,dY,dRotation,dScale;
	boolean animationFadeOut = false;
		

	// ----------------------------------------------------------------------
	// Init Methods
	// ----------------------------------------------------------------------
    public OCVCanvas(int size, Image image, Color color) { 
		edge = size;
        canvasRect = new Rectangle2D.Double(-edge/2.0,-edge/2.0,edge,edge);
		splashImage = image;
		loadingColor = color;
		previousVector = new Vector();
	}
	
	public boolean startDisplayingDocument (OCVDocument aDocument) {
		document = aDocument;
		splashImage = null;
		
		// first find startFrame
		activeFrame = document.startFrame();
		if (activeFrame != null) {
			centerActiveFrame();
		} else {
			if (document.navigationTool() == 1) {
				System.out.println("Error: There is no startFrame!");
				return false;
			}
		}
		
		repaint();
		return true;
	}
	
	
	// ----------------------------------------------------------------------
	// Draw Methods
	// ----------------------------------------------------------------------
	protected void paintComponent(Graphics g) {		// prepare canvas	
		Graphics2D g2 = (Graphics2D)g.create(); // copy g to avoid problems with following paint-methods
		
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		
		if (document == null) { // loading or error
			drawLoadingScreen(g2);
		} else {
			drawDocument(g2);
		}
		g2.dispose();
	}
	
	private void drawLoadingScreen (Graphics2D g2) {
		// background
		g2.setColor(loadingColor); 
		g2.fillRect(0,0,edge,edge);
		
		// splash image
		if (splashImage != null) {
			int height = splashImage.getHeight(this);
			int width = splashImage.getWidth(this);
			
			if ((height != -1) && (width != -1)) {
				g2.drawImage(splashImage,0,0,edge,edge,this);
			}
		}
	}
	
	private void drawDocument(Graphics2D g2) {
		// save original canvas AT
		AffineTransform originalAT = g2.getTransform();
		
		// draw background
		g2.setColor(document.canvasColor()); 
		
		g2.fillRect(0,0,edge,edge);
		
		// if an animation is in progress draw animation
		if (animationCurrentPath != null) {
			switch (animationCurrentPath.transition()) {
				case 2: { //movement
					drawContentOfRect(g2,canvasRect,1,canvasRotation);
					g2.setTransform(originalAT);
					break;
				}
				case 4: { // fade
					Rectangle2D.Double drawRect = drawContentOfRect(g2,canvasRect,1,canvasRotation);
					g2.setTransform(originalAT);
					
					// background
					g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,animationFadeRate));
					g2.setColor(document.canvasColor()); 
					g2.fillRect(0,0,edge,edge);
					g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,1));
					
					
					Point2D.Double center = activeFrame.center();
					double edge = activeFrame.width();
					drawRect = new Rectangle2D.Double(center.x-edge/2,center.y-edge/2,edge,edge);
					drawContentOfRect(g2,drawRect,animationFadeRate,activeFrame.rotation());
					g2.setTransform(originalAT);
					break;
				}
				case 5: { // fade white
					Rectangle2D.Double drawRect = drawContentOfRect(g2,canvasRect,1,canvasRotation);
					g2.setTransform(originalAT);
					g2.setColor(Color.white);
					g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,animationFadeRate));
					g2.fillRect(0,0,edge,edge);
					g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,1));
					break;
				}		
				case 6: { // fade black
					Rectangle2D.Double drawRect = drawContentOfRect(g2,canvasRect,1,canvasRotation);
					g2.setTransform(originalAT);
					g2.setColor(Color.black);
					g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,animationFadeRate));
					g2.fillRect(0,0,edge,edge);
					g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,1));					
					break;
				}		
				case 8: { // dive in
					double frame, fadeRate = animationFadeRate;
					float rotation1, rotation2;
					Rectangle2D.Double clipRect, oldFrameRect, newFrameRect;
					
					if (animationFadeOut) {
						// oldRect
						frame =  canvasRect.width / fadeRate * 0.1;
						oldFrameRect = new Rectangle2D.Double(canvasRect.getCenterX() - frame/2,
															  canvasRect.getCenterY() - frame/2, frame, frame);
						rotation1 = canvasRotation;
						
						// newRect
						Point2D.Double center = activeFrame.center();
						frame = activeFrame.width() / fadeRate;
						newFrameRect = new Rectangle2D.Double(center.x-frame/2, center.y-frame/2, frame, frame);
						rotation2 = activeFrame.rotation();
						
					} else {
						Point2D.Double center = activeFrame.center();
						frame = activeFrame.width() / fadeRate * 0.1;
						oldFrameRect = new Rectangle2D.Double(center.x-frame/2, center.y-frame/2, frame, frame);
						rotation1 = activeFrame.rotation();
						
						frame =  canvasRect.width / fadeRate;
						newFrameRect = new Rectangle2D.Double(canvasRect.getCenterX() - frame/2,
															  canvasRect.getCenterY() - frame/2, frame, frame);
						rotation2 = canvasRotation;
					}
					
					// clipPath
					frame = edge * fadeRate;
					double x = (edge - frame) / 2;
					clipRect = new Rectangle2D.Double(x,x,frame,frame);
					
					
					// -) draw
					drawContentOfRect(g2,oldFrameRect,1,rotation1);
					g2.setTransform(originalAT);
					
					g2.setClip(clipRect);
					g2.setColor(document.canvasColor()); 
					g2.fillRect(0,0,edge,edge);
					drawContentOfRect(g2,newFrameRect,1,rotation2);
					g2.setTransform(originalAT);
					
					break;
					}		
				}
			} else { // draw frame
				drawContentOfRect(g2,canvasRect,1,canvasRotation);
			}
		g2.setTransform(originalAT);
    }
	
	public Rectangle2D.Double drawContentOfRect(Graphics2D g2, Rectangle2D.Double rect, float fraction, float rotation) {
		// 1.: scale/rotate canvas
		double halfSize = edge/2.0;
		double factor = edge/rect.width;
		
		AffineTransform scaleRotateTransform = new AffineTransform();
		scaleRotateTransform.translate(halfSize, halfSize);
		scaleRotateTransform.scale(factor,factor);
		scaleRotateTransform.rotate(-rotation);
		scaleRotateTransform.translate( -halfSize, -halfSize);
		
		// 2.: translate + append scaleRotateTransform
		double xCenter = rect.getCenterX();
		double yCenter = rect.getCenterY();
		
		AffineTransform transform = new AffineTransform();
		transform.translate( -(xCenter-halfSize), -(yCenter-halfSize));
		
		transform.preConcatenate(scaleRotateTransform);
		g2.transform(transform);
		
		// ein rect berechnen, das alle sichtbaren bereiche abdeckt!
		if (canvasRotation != 0) {
			GeneralPath visiblePath = new GeneralPath(rect);
			AffineTransform visibleAT = AffineTransform.getRotateInstance(rotation,xCenter,yCenter);
			Shape visibleShape = visiblePath.createTransformedShape(visibleAT);
			Rectangle2D.Float visibleRect = (Rectangle2D.Float)visibleShape.getBounds2D();
			rect = new Rectangle2D.Double(visibleRect.x,visibleRect.y,visibleRect.width,visibleRect.height);
			
		}
		
		// canvasObjects
		AffineTransform saveAT = g2.getTransform();
		for (Enumeration e = document.canvasObjects().elements() ; e.hasMoreElements() ;){
			OCVObject object = (OCVObject) e.nextElement();
			if (rect.intersects(object.bounds())) 
			{
				object.drawObject(g2,fraction);
				g2.setTransform(saveAT);
			}
		}
		
		return rect;
	}
	
	// ----------------------------------------------------------------------
	// Canvas Functions
	// ----------------------------------------------------------------------
    public Point2D.Double convertPointToView(Point2D.Double aPoint) {
        double x,y;
		
        if (canvasRect.x >= 0) { 
            if (aPoint.x >= 0)
                x = aPoint.x - canvasRect.x; // beide positiv
            else
                x = -(Math.abs(aPoint.x)+canvasRect.x); // canvasRect positiv und aPoint negativ
        } else {
            x = Math.abs(canvasRect.x) + aPoint.x; // beide negativ oder convasRect negativ und aPoint positiv
        }
		
        if (canvasRect.y >= 0) {
            if (aPoint.y >= 0)
                y = aPoint.y - canvasRect.y; // beide positiv
            else
                y = -(Math.abs(aPoint.y)+canvasRect.y); // canvasRect positiv und aPoint negativ
        } else {
            y = Math.abs(canvasRect.y) + aPoint.y; // beide negativ oder convasRect negativ und aPoint positiv
        }
		
        y = edge-y;
        
        return new Point2D.Double(x,y);
    }
	
    public Rectangle2D.Double convertRectToView(Rectangle2D.Double aRect) {
        Point2D.Double newOrigin = convertPointToView(new Point2D.Double(aRect.x,aRect.y));
		
        return new Rectangle2D.Double(newOrigin.x,newOrigin.y-aRect.height,aRect.width,aRect.height);
    }
	
	//public Point convertPointToModel(Point2D aPoint) {
	//	double factor = activeFrame.width() / edge;
     //   return new Point2D.Double(canvasRect.x+aPoint.x*factor,-canvasRect.y+aPoint.y*factor);
    //}
	
   /* public Rectangle2D.Double canvasRect() {
        return canvasRect;
    }
    
    public void resetCanvasRect() {
    }
    
    public void moveCanvasRect(int x, int y) {
        canvasRect.x += x;
        canvasRect.y += y;
        repaint();
    }*/
	
	
	// ----------------------------------------------------------------------
	// Animation Methoden
	// ----------------------------------------------------------------------	
	public void centerActiveFrame() {
		canvasRotation = activeFrame.rotation();
		
		// make sure the startFrame is in the center
		Point2D.Double center = activeFrame.center();
		double objectEdge = activeFrame.width();
		double x = center.x - objectEdge/2.0;
		double y = center.y - objectEdge/2.0;
		canvasRect = new Rectangle2D.Double(x,y,objectEdge,objectEdge);
		
		repaint();
	}
	
	public int calculateAnimationForward(boolean forward) {
		if (document.navigationTool() == 0)
			return 0;
		
		OCVRectObject oldFrame;
		
		if (forward) {
			Vector outgoing = document.outgoingPathsForFrame(activeFrame);
			int count = outgoing.size();
			if (count == 0) {
				return 0;
			} else if (count > 1) {
				Random rnd = new Random(); // init random generator
				int i = Math.abs(rnd.nextInt()) % count; 
				animationCurrentPath = (OCVNavTrailObject)outgoing.elementAt(i);
			} else {
				animationCurrentPath = (OCVNavTrailObject)outgoing.elementAt(0);
			}
			
			oldFrame = activeFrame;
			if (oldFrame.className().equals("OCVActiveComponent"))
				oldFrame = ((OCVActiveComponent)oldFrame).frame();
			previousVector.add(animationCurrentPath);
			activeFrame = document.rectObjectFromID(animationCurrentPath.endRectID());
		} else { // backward
			int c = previousVector.size();
			if (c > 0) {
				oldFrame = activeFrame; // == current frame
				
				animationCurrentPath = (OCVNavTrailObject) previousVector.remove(c-1);
				activeFrame = document.rectObjectFromID(animationCurrentPath.startRectID());
				
				if (activeFrame.className().equals("OCVActiveComponent"))
					activeFrame = ((OCVActiveComponent)activeFrame).frame();
				
			} else {
				return 0; // no previous frame
			}
		}
		
		switch (animationCurrentPath.transition()) {
			case 0: { // none
				return 1;
			}
			case 2: { // movement
				Point2D.Double endPoint, startPoint;
				double width, height;
				double scaleFactor, rotation;
				int segments = animationCurrentPath.segments();
				if (segments < 1)
					return 0;
				
				if (forward) {
					segment = 1; 
					endPoint = animationCurrentPath.endPointOfSegment(segment);
					startPoint = oldFrame.center();
				} else {
					segment = segments;
					if (segment == 1)
						endPoint = activeFrame.center();
					else
						endPoint = animationCurrentPath.endPointOfSegment(segment-1);
					startPoint = animationCurrentPath.endPointOfSegment(segment);
				}
				
				width = endPoint.x - startPoint.x;
				height = endPoint.y - startPoint.y;
				scaleFactor = (int) (activeFrame.width() - oldFrame.width());
				
				rotation = OCVMath.toDegrees(activeFrame.rotation() - oldFrame.rotation());
				if ((rotation > 180) || (rotation < -180)) {
					if (rotation > 0)
						rotation = rotation-360;
					else
						rotation = 360+rotation;
				}
				rotation = OCVMath.toRadians(rotation);
				
				double length = animationFactor = animationCurrentPath.length(document);
				if (length <= Math.abs(scaleFactor)) 
					animationFactor = (int) Math.abs(scaleFactor);
				if (animationFactor <= Math.abs(rotation)) 
					animationFactor = (int)Math.abs(rotation);
				animationFactor /= animationCurrentPath.speed() * 2;
				if (length == 0)
					length = 1;
				
				double seqmentLength = animationCurrentPath.lengthOfSegment(segment, document);
				if (seqmentLength == 0)
					seqmentLength = 1;
				
				times = (int) (animationFactor * (seqmentLength / length)); 
				
				dX = width / times;
				dY = height / times;
				dScale = scaleFactor / animationFactor;
				dRotation = rotation / animationFactor;
				
				break;
			}
			case 4: // fade
			case 5: // fate to white
			case 6: { // fade to black
				animationFadeOut = true;
				setFadeRate(0.1);
				break;
			}
			case 8: {// dive in
				if (forward) {
					animationFadeOut = true;
					setFadeRate(0.1);
				} else {
					animationFadeOut = false;
					setFadeRate(1.0);
				}
			}
		}
		return 2;
	}
	
	public boolean stepAnimation(boolean forward) {
		switch (animationCurrentPath.transition()) {
			case 2: { //movement
				if (times > 0) {
					canvasRect.x += dX-dScale/2.0;
					canvasRect.y += dY-dScale/2.0;
					canvasRect.width += dScale;
					canvasRect.height += dScale;
					canvasRotation += dRotation;
					times--;
				} else {
					Point2D.Double endPoint, startPoint;
					if (forward) {
						segment +=1;
						if (segment <= animationCurrentPath.segments()) {
							endPoint = animationCurrentPath.endPointOfSegment(segment);
							startPoint = animationCurrentPath.endPointOfSegment(segment-1);
						} else {
							return false;
						}
					} else {
						segment -=1;
						if (segment > 0) {
							startPoint = animationCurrentPath.endPointOfSegment(segment);
							if (segment == 1)
								endPoint = activeFrame.center();
							else
								endPoint = animationCurrentPath.endPointOfSegment(segment-1);
						} else {
							return false;
						}
					}
					
					double length = animationCurrentPath.length(document);
					if (length == 0)
						length = 1;
					
					double seqmentLength = animationCurrentPath.lengthOfSegment(segment, document);
					if (seqmentLength == 0)
						seqmentLength = 1;
					
					times = (int) (animationFactor * (seqmentLength / length)); 
					
					dX = (endPoint.x - startPoint.x) / times;
					dY = (endPoint.y - startPoint.y) / times;
				}
				return true;
			}
			case 4: { // fade
				setFadeRate(animationFadeRate+(0.005*animationCurrentPath.speed() * 2));
				// 1/2 slower because ther is no fade in 0.01/2=0.005
				if (animationFadeRate >= 1.0) {
					return false;
				} 
				return true;
			}
			case 8: { // dive in
				if (forward) {
					setFadeRate(animationFadeRate+(0.005*animationCurrentPath.speed() * 2));
					// 1/2 slower because ther is no fade in 0.01/2=0.005
					if (animationFadeRate >= 1.0) {
						return false;
					} 
				} else{
					setFadeRate(animationFadeRate-(0.005*animationCurrentPath.speed() * 2));
					// 1/2 slower because ther is no fade in 0.01/2=0.005
					if (animationFadeRate <= 0.1) {
						return false;
					} 
				}
				return true;	
			}
				
			case 5: // fade white
			case 6: { // fade black
				if (animationFadeOut) {
					if (animationFadeRate >= 1.0) {
						// make sure the new frame is in the center.
						centerActiveFrame();
						animationFadeOut = false;
					} else {
						setFadeRate(animationFadeRate+(0.01*animationCurrentPath.speed() * 2));
					}
				} else {
					if (animationFadeRate <= 0.0) {
						return false;
					} else {
						setFadeRate(animationFadeRate-(0.01*animationCurrentPath.speed() * 2));
					}
				}
				return true;
			}
		}
		return true;
	}
	
	public void finishAnimation() {
		animationCurrentPath = null;
	}
		
	public boolean validPreviousFrame() {
		if (previousVector.size() > 0) {
			return true;
		} else {
			return false;
		}
	}
	
	public boolean validNextFrame() {
		if ((document == null) || (activeFrame == null))
			return false;
		
		if (document.outgoingPathsForFrame(activeFrame).size() > 0) {
			return true;
		} else {
			return false;
		}
	}
	
	public boolean waitForInput() {
		return ((OCVNavFrameObject)activeFrame).waiting();
	}
	
	private void setFadeRate(double rate) {
		animationFadeRate = (float) rate;
		if (animationFadeRate > 1.0)
			animationFadeRate = 1;
		else if (animationFadeRate < 0.0)
			animationFadeRate = 0;
	}
	
	public double scaleFactor() {
		return canvasRect.width / edge;
	}
	
	
	public boolean activeComponentHit(Point point, boolean select) {
		OCVRectObject tempFrame = (OCVRectObject) ((OCVNavFrameObject)activeFrame).activeComponentHitTest(point);
		if ((tempFrame != null) && (document.outgoingPathsForFrame(tempFrame).size() > 0)) {
			if (select)
				activeFrame = tempFrame;
			return true;
		} else {
			return false;
		}
	}


	// ----------------------------------------------------------------------
	// Free Navigation Methoden
	// ----------------------------------------------------------------------	
	public void moveXY(double x, double y) {
		canvasRect.x += Math.cos(canvasRotation) * x - Math.sin(canvasRotation) * y;
		canvasRect.y += Math.cos(canvasRotation) * y + Math.sin(canvasRotation) * x;		
		repaint();
	}

	public void rotateCW() {
		canvasRotation =  (float) (canvasRotation-OCVMath.toRadians(3));
		repaint();
	}

	public void rotateCCW() {
		canvasRotation =  (float) (canvasRotation+OCVMath.toRadians(3));
		repaint();
	}

	public void zoomIn() {
		double newEdge = canvasRect.width / 1.05;
		double shift = (canvasRect.width - newEdge) / 2.0;
		canvasRect = new Rectangle2D.Double(canvasRect.x+shift,canvasRect.y+shift,newEdge,newEdge);
		repaint();
	}

	public void zoomOut() {
		double newEdge = canvasRect.width * 1.05;
		double shift =  (newEdge - canvasRect.width) / 2.0;
		canvasRect = new Rectangle2D.Double(canvasRect.x-shift,canvasRect.y-shift,newEdge,newEdge);
		repaint();
	}

	public void resetFreeNavigation() {
		canvasRotation = 0;
		if (activeFrame != null) {
			centerActiveFrame(); // does repaint
		} else {	
			canvasRect = new Rectangle2D.Double(-edge/2.0,-edge/2.0,edge,edge);
			repaint();
		}
	}
}
